 #include <ESP8266WiFi.h>
#include <SPI.h> 
#include <EEPROM.h>

WiFiServer server(80);
//WLAN
const char* ssid = "myWlanName";
const char* password = "myPassword";
const String ver = "V 09.02 20170816";
/*
ChangeLog: 
  - waiting time after reading eeprom set to 15 sec because before not stable connect to wlan
  - set volume after set mute on/off
*/
uint8_t MAC_array[6];
char MAC_char[18];
//MAC:  5C:CF:7F:99:93:E4 (Test-ESP)//Hardware - Relays on D1...D4, SlaveSelect on D0 (! not D8!)
//SCLK on D5, MOSI (SDI) on D7 must not declared
const byte ch1Pin = D1; const byte ch2Pin = D2; const byte ch3Pin = D3; const byte ch4Pin = D4; 
const int slaveSelectPin = D0;
const int ledPin = D8;
//EEPROM with example and configuration
const int storageLength = 60;
char storage[storageLength] = "1100110120130Kanal1    Kanal2    Kanal3    Kanal4    ";
String name1 = "channel1"; String name2 = "channel2"; String name3 = "channel3"; String name4 = "channel4";
byte value1 = 150; byte value2 = 150; byte value3 = 150; byte value4 = 150; 
//HTML and control
String sResponse, sHeader;
String sPath = "", sParam = "", sCmd = "";
boolean isNewCmd = false;
String oldPath = sPath; String oldParam = sParam;
byte volume = 150; byte saveVolume = volume;
byte channel = 0; //0=all off, 1=channel 1, 2=channel 2...
byte oldChannel = channel;
String clientType = ""; String calledSite = ""; byte startChannel = 1;

//*************************************************************************************************************
//**********************  Start with setup()
//*************************************************************************************************************

void setup() 
{
  // SPI
  pinMode(slaveSelectPin, OUTPUT);
  SPI.begin(); 

  // Serial
  Serial.begin(115200);
  delay(10);

/*
  WiFi.macAddress(MAC_array);
  for (int i = 0; i < sizeof(MAC_array); ++i){
    sprintf(MAC_char,"%s%02x:",MAC_char,MAC_array[i]);
  }
  Serial.println("");
  Serial.println(MAC_char);
*/
  //EEPROM
  //writeConfig2EEPROM();
  readConfigFromEEPROM();
  //var2storage();

  //Pins for relays
  pinMode(ch1Pin, OUTPUT); pinMode(ch2Pin, OUTPUT); pinMode(ch3Pin, OUTPUT); pinMode(ch4Pin, OUTPUT);
  digitalWrite(ch1Pin, LOW); digitalWrite(ch2Pin, LOW); digitalWrite(ch3Pin, LOW); digitalWrite(ch4Pin, LOW);
  pinMode(ledPin, OUTPUT); digitalWrite(ledPin, LOW); //wlan-led aus
  delay(100);
  
  //set Startparameters from Config
  setStartparametersFromConfig();
  calledSite = "main";

  // inital connect for WIFI
  WiFi.mode(WIFI_STA);
  WiFiStart();
  delay(100);
}

//*************************************************************************************************************
//**********************  and Main program
//*************************************************************************************************************

void loop() 
{
  // if not connected --> return and start again
  if (WiFi.status() != WL_CONNECTED) {
    digitalWrite(ledPin, LOW);
    WiFiStart();
  } else {
    digitalWrite(ledPin, HIGH);
  }

  // if not client connected --> return and again
  WiFiClient client = server.available();
  if (!client) {
    return;
  }
  
  //check timeout --> for client-connection timeout
  unsigned long ultimeout = millis()+250;
  while(!client.available() && (millis()<ultimeout)) {
    delay(1);
  }
  if(millis()>ultimeout){
    return; 
  }
  
  String sRequest = client.readStringUntil('\r');
  client.flush();
  if(sRequest=="") {
    client.stop();
    return;
  }
  
  // get path; end of path is either space or ?
  // Syntax is e.g. GET /?channel=1 HTTP/1.1
  isNewCmd = false;
  String sGetstart="GET /?"; //Start of the GET-String
  int iStart, iEndSpace, iEndQuest;
  iStart = sRequest.indexOf(sGetstart); 
  if (iStart>=0) {    
    //GET found
    iStart +=+ sGetstart.length(); 
    iEndSpace = sRequest.indexOf(" ",iStart); //Ende des zu verwertenden Strings
    iEndQuest = sRequest.indexOf("=",iStart); //= als Trenner zwischen Parameter und Cmd
    sPath  = sRequest.substring(iStart,iEndQuest);
    sParam = sRequest.substring(iEndQuest + 1,iEndSpace);
    delay(50);
    checkCommands();
  }
  
  //generating HTML response
  // 404 for non-matching path
  if(!(
      (sPath == "volume") || 
      (sPath == "mute") || 
      (sPath == "channel") || 
      (sPath == "type") || 
      (sPath == "") || 
      (sPath == "type") || 
      (sPath == "call") ||
      (sPath == "name1") ||
      (sPath == "name2") ||
      (sPath == "name3") ||
      (sPath == "name4") ||
      (sPath == "startch")
      )) 
  {
      if (clientType != "ANDROID") {
        //Serial.println("404 Not found");
        sResponse="<html><head><title>404 Not Found</title></head><body><h1>404 Not Found</h1><p>The requested URL was not found on this server.</p></body></html>";
      }
      else
      {
        sResponse = "unknown command";
      }
      sHeader  = "HTTP/1.1 404 Not found\r\n";
      sHeader += "Content-Length: ";
      sHeader += sResponse.length();
      sHeader += "\r\n";
      sHeader += "Content-Type: text/html\r\n";
      sHeader += "Connection: close\r\n";
      sHeader += "\r\n";
  }
  else
  {
    if (clientType != "ANDROID") {
      if (calledSite == "main") {
        makeHtmlMain();
      }
      else {
        makeHtmlSettings();
      }
    }
    else {
      //response for Android
      sResponse = "Command: " + sPath + "\r\n";
      sResponse += "Value: " + sParam;
    }
    sHeader  = "HTTP/1.1 200 OK\r\n";
    sHeader += "Content-Length: ";
    sHeader += sResponse.length();
    sHeader += "\r\n";
    sHeader += "Content-Type: text/html\r\n";
    sHeader += "Connection: close\r\n";
    sHeader += "\r\n";
  }
  
  // Send the response to the client
  client.print(sHeader);
  client.print(sResponse);
  
  // and stop the client
  client.stop();
}


//*************************************************************************************************************
//**********************  all functions for special purpose
//*************************************************************************************************************

void saveConfig() {
  var2storage();
  writeConfig2EEPROM();
}

void writeConfig2EEPROM() {
  //writes the data in char array storage to EEPROM
  EEPROM.begin(512);
  delay(10);
  for (int i = 0; i < storageLength; i++) {
    EEPROM.write(i, storage[i]);
  }
  EEPROM.commit();
  delay(500);
}

void var2storage() {
  //building the char array storage
  String tmp = String(startChannel);
  String tmp1 = String(value1);
  tmp += fillCharacters(tmp1, 3);
  tmp1 = String(value2);
  tmp += fillCharacters(tmp1, 3);
  tmp1 = String(value3);
  tmp += fillCharacters(tmp1, 3);
  tmp1 = String(value4);
  tmp += fillCharacters(tmp1, 3);
  tmp += fillCharacters(name1, 10);
  tmp += fillCharacters(name2, 10);
  tmp += fillCharacters(name3, 10);
  tmp += fillCharacters(name4, 10);
  while (tmp.length() < storageLength) {
    tmp = tmp + " ";
  }
  tmp.toCharArray(storage, storageLength);
}

String fillCharacters(String s, int count) {
  //only to get fixed field length for fields in char array
  s.trim();
  String result = s;
  while(result.length() < count) {
     result = " " + result;
  }
  return result;
}

void readConfigFromEEPROM()
{
  //get data from EEPROM and store to variables
  EEPROM.begin(512);
  delay(50);
  String tmp = "";  
  for (int i = 0; i < storageLength; i++)
  {
    storage[i] = EEPROM.read(i);
    tmp = tmp + storage[i]; 
  }
  startChannel = tmp.substring(0, 1).toInt();
  value1 = tmp.substring(1, 4).toInt();
  value2 = tmp.substring(4, 7).toInt();
  value3 = tmp.substring(7, 10).toInt();
  value4 = tmp.substring(10, 13).toInt();
  name1 = tmp.substring(13, 23);
  name1.trim();
  name2 = tmp.substring(23, 33);
  name2.trim();
  name3 = tmp.substring(33, 43);
  name3.trim();
  name4 = tmp.substring(43, 53);
  name4.trim();
  delay(15000);
}

String getFirstValue(String s) {
  //extract first value from save config, used in case of sending settings from client (name and volume of one channel)
  // example: s = Name1&value1=13 ---
  int i = s.indexOf("&");
  String result = s.substring(0, i);
  result = formatResult(result);
  return result; 
}

String getSecondValue(String s) {
  //extract second value from save config
  int i = s.indexOf("=");
  String result = s.substring(i + 1);
  return result;
}

String formatResult(String s) {
  //formatting Name from browser
  if (s.length() > 10) {
    s = s.substring(0, 10);
  }
  s.replace("+", " ");
  return s;
}

void setStartparametersFromConfig() {
  //set channel after power on
  switch (startChannel) {
    case 1:   digitalWrite(ch1Pin, HIGH);
      volume = value1;
      channel = 1;
      break;
    case 2:   digitalWrite(ch2Pin, HIGH);
      volume = value2;
      channel = 2;
      break;
    case 3:   digitalWrite(ch3Pin, HIGH);
      volume = value3;
      channel = 3;
      break;
    case 4:   digitalWrite(ch4Pin, HIGH);
      volume = value4;
      channel = 4;
      break;
  }
  pga2310Write(volume);
}

void makeHtmlMain() {
  //make the main HTML-file 
  String s = "not connected";
  switch (channel) {
    case 1: s = name1;
      break;
    case 2: s = name2;
      break;
    case 3: s = name3;
      break;
    case 4: s = name4;
      break;
  }
  sResponse = "<html><head><title>Controlling PGA2310 with ESP8266</title>";
  sResponse += "<meta name=\"viewport\" content=\"width=device-width, initial-scale=1.0, user-scalable=yes\"></head>";
  sResponse += "<body bgcolor=\"#663333\" text=\"#FFCC99\">";
  sResponse += "<h3>Controlling PGA2310 with ESP8266</h3>";
  sResponse += "<hr noshade width=90% align=left color=\"#FF0000\">";
  sResponse += "<table>"; //<tr><td>&nbsp;</td></tr>";
   sResponse += "<tr><td>Channel:</td><td>" + s + "</td></tr>";
  sResponse += "<tr><td>Volume:</td><td>" + String(volume) + "</td><td>(0 ... 255)</td></tr>";
  sResponse += "</table>";
  sResponse += "<hr noshade width=90% align=left color=\"#FF0000\">";
  sResponse += "<table><tr><td>Set volume</td><td><a href=\"?volume=plus\"><button style=\"width:8em\">+1</button></td>";
  sResponse += "<td><a href=\"?volume=minus\"><button style=\"width:8em\">-1</button></a></td></tr>";
  sResponse += "<tr><td></td><td><a href=\"?volume=PLUS\"><button style=\"width:8em\">+5</button style=\"width:8em\"></td><td><a href=\"?volume=MINUS\"><button style=\"width:8em\">-5</button></a></td></tr>";
  sResponse += "<tr><td>&nbsp;</td></tr>";
  sResponse += "<tr><td>Mute:</td><td><a href=\"?mute=1\"><button style=\"width:8em\">ON</button style=\"width:8em\"></a></td><td><a href=\"?mute=0\"><button style=\"width:8em\">OFF</button></a></td></tr>";
  sResponse += "<tr><td>&nbsp;</td></tr>";
  sResponse += "<tr><td>Channel:</td><td><a href=\"?channel=0\"><button style=\"width:8em\">All off</button></a></td></tr>";
  sResponse += "<tr><td>" + String(channel) + "</td><td><a href=\"?channel=1\"><button style=\"width:8em\">" + name1 + "</button></a></td>";
  sResponse += "<td><a href=\"?channel=2\"><button style=\"width:8em\">" + name2 + "</button></a></td></tr>";
  sResponse += "<tr><td></td><td><a href=\"?channel=3\"><button style=\"width:8em\">" + name3 + "</button></a></td>";
  sResponse += "<td><a href=\"?channel=4\"><button style=\"width:8em\">" + name4 + "</button></a></td></tr>";
  sResponse += "</table>"; 
  sResponse += "<hr noshade width=90% align=left color=\"#FF0000>\">";
  sResponse += "<p><a href=\"?type=HTML\"><button>Switch from App to Browser</button></a>";
  sResponse += "<p><a href=\"?call=settings\"><button style=\"width:8em\">Settings</button></a>";
  sResponse += "<hr noshade width=90% align=left color=\"#FF0000\">";
  sResponse += ver + "</body></html>";
}

void makeHtmlSettings() {
  sResponse = "<html><head><title>Settings for Volume Control</title>";
  sResponse += "<meta name=\"viewport\" content=\"width=device-width, initial-scale=1.0, user-scalable=yes\"></head>";
  sResponse += "<body bgcolor=\"#663333\" text=\"#FFCC99\">";
  sResponse += "<h3>Settings for Volume Control</h3>";
  sResponse += "<hr noshade width=90% align=left color=\"#FF0000\">";
  sResponse += "<table><form>";
  sResponse += "<tr><td>Start with</td><td><input name=\"startch\"></td><td>&nbsp;&nbsp;&nbsp;" + String(startChannel) + "</td></tr>";
  sResponse += "<tr><td>Channel:</td><td><input type=\"submit\" value=\"Set\"></td></tr>";
  sResponse += "<tr><td>&nbsp;<td></tr></form>";
  sResponse += "<form><tr><td>Channel 1:</td></tr>";
  sResponse += "<tr><td>Name:</td><td><input name=\"name1\"></td><td>&nbsp;&nbsp;&nbsp;" + name1 + "</td></tr>";
  sResponse += "<tr><td>Start value:</td><td><input name=\"value1\"></td><td>&nbsp;&nbsp;&nbsp;" + String(value1) + "</td></tr>";
  sResponse += "<tr><td>&nbsp;</td><td><input type=\"submit\" value=\"Set Channel 1\"></td></tr>";
  sResponse += "</form><form>";
  sResponse += "<tr><td>Channel 2:</td></tr>";
  sResponse += "<tr><td>Name:</td><td><input name=\"name2\"></td><td>&nbsp;&nbsp;&nbsp;" + name2 + "</td></tr>";
  sResponse += "<tr><td>Start value:</td><td><input name=\"value2\"></td><td>&nbsp;&nbsp;&nbsp;" + String(value2) + "</td></tr>";
  sResponse += "<tr><td>&nbsp;</td><td><input type=\"submit\" value=\"Set Channel 2\"></td></tr>";
  sResponse += "</form><form>";
  sResponse += "<tr><td>Channel 3:</td></tr>";
  sResponse += "<tr><td>Name:</td><td><input name=\"name3\"></td><td>&nbsp;&nbsp;&nbsp;" + name3 + "</td></tr>";
  sResponse += "<tr><td>Start value:</td><td><input name=\"value3\"></td><td>&nbsp;&nbsp;&nbsp;" + String(value3) + "</td></tr>";
  sResponse += "<tr><td>&nbsp;</td><td><input type=\"submit\" value=\"Set Channel 3\"></td></tr>";
  sResponse += "</form><form>";
  sResponse += "<tr><td>Channel 4:</td></tr>";
  sResponse += "<tr><td>Name:</td><td><input name=\"name4\"></td><td>&nbsp;&nbsp;&nbsp;" + name4 + "</td></tr>";
  sResponse += "<tr><td>Start value:</td><td><input name=\"value4\"></td><td>&nbsp;&nbsp;&nbsp;" + String(value4) + "</td></tr>";
  sResponse += "<tr><td>&nbsp;</td><td><input type=\"submit\" value=\"Set Channel 4\"></td></tr>";
  sResponse += "<tr><td>&nbsp;</td></tr></form>";
  sResponse += "</table>";
  sResponse += "<hr noshade width=90% align=left color=\"#FF0000\">";
  sResponse += "<p><a href=\"?call=main\"><button style=\"width:8em\">Back</button></a>";
  sResponse += "<hr noshade width=90% align=left color=\"#FF0000\">";
  sResponse += "</body></html>";
}

void checkCommands() {
  //********** all actions to react for commands from client
  //first all functions for save configuration ********** 
  Serial.println("---" + sPath + "---" + sParam + "---");
  int i = 0;
  if (sPath == "startch") {
    startChannel = sParam.toInt();
    saveConfig();
  }
  if (sPath == "name1"){
    name1 = getFirstValue(sParam);
    i = getSecondValue(sParam).toInt();
    if (i > 0) {
      if (i <= 255) {
            value1 = i;
      }
    }
    saveConfig();
  }
  if (sPath == "name2"){
    name2 = getFirstValue(sParam);
    i = getSecondValue(sParam).toInt();
    if (i > 0) {
      if (i <= 255) {
            value2 = i;
      }
    }
    saveConfig();
  }
  if (sPath == "name3"){
    name3 = getFirstValue(sParam);
    i = getSecondValue(sParam).toInt();
    if (i > 0) {
      if (i <= 255) {
            value3 = i;
      }
    }
    saveConfig();
  }
  if (sPath == "name4"){
    name4 = getFirstValue(sParam);
    i = getSecondValue(sParam).toInt();
    if (i > 0) {
      if (i <= 255) {
            value4 = i;
      }
    }
    saveConfig();
  }
  //********** all functions for save configuration ********** 
  if (sPath == "type") {
    //********** switch from ANDROID to HTML (generate HTML for Browser) ********** 
    clientType = sParam;
  }
  if (sPath == "call") {
    //********** switch from main to settings and back and generate the correspondig html-page ********** 
    calledSite = sParam;
  }
  if (sPath == "volume"){
    //********** all functions for setting volume in steps of 1 or 5 (plus/PLUS) ********** 
    if (sParam == "plus") {
      if (volume < 255) {
        volume++;
      }
    }
    if (sParam == "minus") {
      if (volume > 0) {
        volume--;
      }
    }
    if (sParam == "PLUS") {
      if (volume < 251) {
        volume = volume + 5;
      }
    }
    if (sParam == "MINUS") {
      if (volume > 5) {
        volume = volume - 5;
      }
    }
    saveVolume = volume;
    pga2310Write(volume);
  }
  if (sPath == "mute") {
    if (sParam.toInt() == 0) {
      volume = saveVolume;
      pga2310Write(volume);
    }
    if (sParam.toInt() == 1) {
      volume = 0;
      pga2310Write(volume);
    }
  }
  if (sPath == "channel") {
    if ((sParam.toInt() <= 4) && (sParam.toInt() >= 0)) {
      channel = sParam.toInt();
    }
    if (channel != oldChannel) {
      oldChannel = channel;
      digitalWrite(ch1Pin, LOW); digitalWrite(ch2Pin, LOW); digitalWrite(ch3Pin, LOW); digitalWrite(ch4Pin, LOW); 
      delay(100);   //debounce
      switch (channel) {
        case 1:
          digitalWrite(ch1Pin, HIGH);
          break;
        case 2:
          digitalWrite(ch2Pin, HIGH);
          break;
         case 3:
          digitalWrite(ch3Pin, HIGH);
          break;
        case 4:
          digitalWrite(ch4Pin, HIGH);
          break;
      }
    }
  }
}

void pga2310Write(int value) {
  //send data to PGA23xx
  digitalWrite(slaveSelectPin, LOW);
  SPI.transfer(value);
  SPI.transfer(value);
  digitalWrite(slaveSelectPin, HIGH);
}

void WiFiStart()
{
  boolean ledOn = false;
  digitalWrite(ledPin, LOW);
  Serial.println();
  Serial.print("Connecting to ");
  Serial.println(ssid);
  
  WiFi.begin(ssid, password);
  
  while (WiFi.status() != WL_CONNECTED) {
    if (ledOn) {
      ledOn = false; 
    } else {
      ledOn = true;
    }
    if (ledOn) {
      digitalWrite(ledPin, HIGH);
    } else {
      digitalWrite(ledPin, LOW);
    }
    
    delay(500);
    Serial.print(".");
  }
  Serial.println("");
  Serial.println("WiFi connected");
  
  server.begin();
  Serial.println("Server started");
  Serial.println(WiFi.localIP());
}


